
gaze_data_fname                 = 'gazeData.tsv'
frame_timestamps_fname          = 'frameTimestamps.tsv'
scene_camera_video_fname_stem   = 'worldCamera'
scene_camera_calibration_fname  = 'calibration.xml'
gaze_overlay_video_file         = 'gazeOverlay.mp4'
marker_pose_prefix              = 'markerPose_'
fixation_classification_prefix  = 'fixation_classification'
fixation_assignment_prefix      = 'fixation_assignment'
validation_offset_fname         = 'dataQuality.tsv'


"""
Copy data already in common format into a glassesTools recording.
Name of recording will be the name of the folder that is imported.
"""

import shutil
import pathlib
import pandas as pd

from ..recording import Recording
from ..eyetracker import EyeTracker
from .. import gaze_headref
from .. import naming, video_utils

def importData(output_dir: str|pathlib.Path=None, source_dir: str|pathlib.Path=None, rec_info: Recording=None, device_name: str=None, copy_scene_video = True, source_dir_as_relative_path = False, cam_cal_file: str|pathlib.Path=None) -> Recording:
    from . import check_folders, _store_data
    output_dir, source_dir, rec_info, device_name = check_folders(output_dir, source_dir, rec_info, EyeTracker.Generic, device_name)
    print(f'processing: {source_dir.name} -> {output_dir}')

    ### check and copy needed files to the output directory
    print('  Check and copy raw data...')
    ### check recording and get export directory
    if rec_info is not None:
        checkRecording(source_dir, rec_info, device_name)
    else:
        rec_info = getRecordingInfo(source_dir, device_name)
        if rec_info is None:
            raise RuntimeError(f"The folder {source_dir} is not recognized as a {EyeTracker.Generic.value} recording.")

    # make output dir
    if not output_dir.is_dir():
        output_dir.mkdir()


    ### copy the raw data to the output directory
    srcVid, destVid, gotFrameTs, gotCal = copyGenericRecording(source_dir, output_dir, copy_scene_video, cam_cal_file)
    if destVid:
        rec_info.scene_video_file = destVid.name
    else:
        rec_info.scene_video_file =  srcVid.name

    if not gotFrameTs:
        frameTimestamps = video_utils.get_frame_timestamps_from_video(rec_info.get_scene_video_path())
    else:
        frameTimestamps = None

    if not gotCal:
        print('    !! No camera calibration provided!')

    if not rec_info.duration:
        # if duration not known, fill it
        # make a reasonable estimate of duration
        gaze = gaze_headref.read_dict_from_file(output_dir/naming.gaze_data_fname)[0]
        framets = frameTimestamps   # local copy to not accidentally trigger any overwriting in _store_data() below
        if framets is None:
            framets = pd.read_csv(output_dir/naming.frame_timestamps_fname, delimiter='\t', index_col='frame_idx')
        gt0 = gaze[min(gaze.keys())][0].timestamp_ori
        gte = gaze[max(gaze.keys())][-1].timestamp_ori
        rec_info.duration = float(round(max(gte-gt0,framets.timestamp.iat[-1])))

    _store_data(output_dir, None, frameTimestamps, rec_info, source_dir_as_relative_path=source_dir_as_relative_path)

    return rec_info

def getRecordingInfo(inputDir: str|pathlib.Path, device_name: str=None) -> Recording:
    # returns None if not a recording directory
    inputDir = pathlib.Path(inputDir)

    rec_info_fname = inputDir/Recording.default_json_file_name
    if rec_info_fname.is_file():
        recInfo = Recording.load_from_json(rec_info_fname)
        if recInfo.eye_tracker!=EyeTracker.Generic:
            print(f"A recording for a \"{EyeTracker.Generic.value}\" eye tracker was not found in the folder {inputDir}.")
            return None
        if recInfo.eye_tracker_name!=device_name:
            print(f"A recording for a \"{device_name}\" device was not found in the folder {inputDir}.")
            return None
        # override inputDir to make sure its set correctly
        recInfo.source_directory = inputDir
    else:
        recInfo = Recording(source_directory=inputDir, eye_tracker=EyeTracker.Generic, eye_tracker_name=device_name)
        # get recording info
        recInfo.name = inputDir.name

    # check expected files are present
    for f in ('worldCamera.mp4','gazeData.tsv'):
        if not (inputDir/f).is_file():
            print(f'This directory does not contain a valid generic recording for a {device_name} eye tracker. The {f} file is not found in the input directory {inputDir}.')
            return None

    # we got a valid recording
    # return what we've got
    return recInfo

def checkRecording(inputDir: str|pathlib.Path, recInfo: Recording, device_name: str=None):
    actualRecInfo = getRecordingInfo(inputDir, device_name)

    if actualRecInfo is None or recInfo.name!=actualRecInfo.name:
        raise ValueError(f"A recording with the name \"{recInfo.name}\" was not found in the folder {inputDir}.")

    # make sure caller did not mess with recInfo
    if recInfo.eye_tracker_name!=actualRecInfo.eye_tracker_name:
        raise ValueError(f"A recording for a \"{recInfo.eye_tracker_name}\" device was not found in the folder {inputDir}.")

def copyGenericRecording(inputDir: pathlib.Path, outputDir: pathlib.Path, copy_scene_video:bool, cam_cal_file: str|pathlib.Path|None):
    gazeFile = inputDir/'gazeData.tsv'
    if not gazeFile.is_file():
        raise RuntimeError(f'The {gazeFile} file is not found in the input directory {inputDir}')
    shutil.copy2(gazeFile, outputDir / naming.gaze_data_fname)

    vidSrcFile = inputDir/'worldCamera.mp4'
    if not vidSrcFile.is_file():
        raise RuntimeError(f'The {vidSrcFile} file is not found in the input directory {inputDir}')
    if copy_scene_video:
        vidDestFile = outputDir / f'{naming.scene_camera_video_fname_stem}.mp4'
        shutil.copy2(vidSrcFile, vidDestFile)
    else:
        vidDestFile = None

    frame_ts_file = inputDir/'frameTimestamps.tsv'
    gotFrameTs = frame_ts_file.is_file()
    if gotFrameTs:
        shutil.copy2(frame_ts_file, outputDir / naming.frame_timestamps_fname)

    if cam_cal_file is not None:
        shutil.copy2(cam_cal_file, outputDir / naming.scene_camera_calibration_fname)
        gotCal = True
    else:
        cal_file = inputDir/'calibration.xml'
        gotCal = cal_file.is_file()
        if gotCal:
            shutil.copy2(cal_file, outputDir / naming.scene_camera_calibration_fname)

    return vidSrcFile, vidDestFile, gotFrameTs, gotCal


import shutil
import pathlib
import json
import pandas as pd

from ..recording import Recording
from ..eyetracker import EyeTracker
from .. import naming, timestamps, video_utils

def importData(output_dir: str|pathlib.Path=None, source_dir: str|pathlib.Path=None, rec_info: Recording=None, copy_scene_video = True, source_dir_as_relative_path = False) -> Recording:
    from . import check_folders, _store_data
    output_dir, source_dir, rec_info, _ = check_folders(output_dir, source_dir, rec_info, EyeTracker.Meta_Aria_Gen_1)
    print(f'processing: {source_dir.name} -> {output_dir}')

    ### check and copy needed files to the output directory
    print('  Check and copy raw data...')
    ### check recording and get export directory
    if rec_info is not None:
        checkRecording(source_dir, rec_info)
    else:
        rec_info = getRecordingInfo(source_dir)
        if rec_info is None:
            raise RuntimeError(f"The folder {source_dir} is not recognized as a {EyeTracker.Meta_Aria_Gen_1.value} recording.")

    # make output dir
    if not output_dir.is_dir():
        output_dir.mkdir()


    ### copy the raw data to the output directory
    srcVid, destVid, gazeData, frameTimestamps = copyRecording(source_dir, output_dir, rec_info, copy_scene_video)
    if destVid:
        rec_info.scene_video_file = destVid.name
    else:
        rec_info.scene_video_file =  srcVid.name

    _store_data(output_dir, gazeData, frameTimestamps, rec_info, source_dir_as_relative_path=source_dir_as_relative_path)

    return rec_info

def getRecordingInfo(inputDir: str|pathlib.Path) -> Recording:
    # returns None if not a recording directory
    inputDir = pathlib.Path(inputDir)
    recInfo = Recording(source_directory=inputDir, eye_tracker=EyeTracker.Meta_Aria_Gen_1)

    # check expected files are present
    for f in ('metadata.json','worldCamera.mp4','gaze.tsv','calibration.xml'):
        if not (inputDir/f).is_file():
            print(f'This directory does not contain a valid {EyeTracker.Meta_Aria_Gen_1.value} recording export. The {f} file is not found in the input directory {inputDir}. Make sure you run the meta_aria_gen1_exporter.py script on the recording\'s vrs file')
            return None

    with open(inputDir/'metadata.json', 'r') as f:
        metadata = json.load(f)
    recInfo.name = metadata['name']
    recInfo.scene_camera_serial = metadata['scene_camera_serial']
    recInfo.duration = int(metadata['duration']/1000)   # in us, convert to ms
    recInfo.glasses_serial = metadata['glasses_serial']
    recInfo.start_time = timestamps.Timestamp(metadata['start_time'])
    recInfo.scene_video_file = 'worldCamera.mp4'

    # we got a valid recording
    # return what we've got
    return recInfo

def checkRecording(inputDir: str|pathlib.Path, recInfo: Recording):
    actualRecInfo = getRecordingInfo(inputDir)

    if actualRecInfo is None or recInfo.name!=actualRecInfo.name:
        raise ValueError(f"A recording with the name \"{recInfo.name}\" was not found in the folder {inputDir}.")

    # make sure caller did not mess with recInfo
    if recInfo.eye_tracker!=actualRecInfo.eye_tracker:
        raise ValueError(f"A recording for a \"{recInfo.eye_tracker.value}\" device was not found in the folder {inputDir}.")

def copyRecording(inputDir: pathlib.Path, outputDir: pathlib.Path, rec_info: Recording, copy_scene_video:bool):
    gazeFile = inputDir/'gaze.tsv'
    if not gazeFile.is_file():
        raise RuntimeError(f'The {gazeFile} file is not found in the input directory {inputDir}')
    gazeData = pd.read_csv(gazeFile,sep='\t',index_col='timestamp')

    vidSrcFile = inputDir/'worldCamera.mp4'
    if not vidSrcFile.is_file():
        raise RuntimeError(f'The {vidSrcFile} file is not found in the input directory {inputDir}')
    if copy_scene_video:
        vidDestFile = outputDir / f'{naming.scene_camera_video_fname_stem}.mp4'
        shutil.copy2(vidSrcFile, vidDestFile)
    else:
        vidDestFile = None

    # get video timestamps
    frameTimestamps = video_utils.get_frame_timestamps_from_video(rec_info.get_scene_video_path())

    # convert gaze timestamps from us to ms
    gazeData = gazeData.reset_index()
    gazeData.loc[:,'timestamp'] /= 1000.0
    # add frame_idx for each gaze sample
    frameIdx = video_utils.timestamps_to_frame_number(gazeData.loc[:,'timestamp'].values,frameTimestamps['timestamp'].values)
    gazeData.insert(1,'frame_idx',frameIdx['frame_idx'].values)
    gazeData = gazeData.set_index('timestamp')

    cal_file = inputDir/'calibration.xml'
    if not cal_file.is_file():
        raise RuntimeError(f'The {cal_file} file is not found in the input directory {inputDir}')
    shutil.copy2(cal_file, outputDir / naming.scene_camera_calibration_fname)

    return vidSrcFile, vidDestFile, gazeData, frameTimestamps


import orjson
import json
import typing
import dataclasses
import pathlib



def json_encoder(obj):
    matches: dict[typing.Type, TypeEntry] = {}
    for t in TYPE_REGISTRY:
        if isinstance(obj, t.type):
            matches[t.type] = t
    if matches:
        # find most precise match and use that
        t = None
        for t2 in matches:
            if t is None or issubclass(t2,t):
                t = t2
        return {matches[t].reg_name: matches[t].to_json(obj)}
    else:
        raise TypeError(f'type {type(obj)} cannot be serialized')

def dump(obj, file: pathlib.Path):
    data = orjson.dumps(obj, json_encoder, orjson.OPT_INDENT_2|orjson.OPT_PASSTHROUGH_SUBCLASS)
    with open(file, 'wb') as f:
        f.write(data)


def json_decoder(d):
    for t in TYPE_REGISTRY:
        if t.reg_name in d:
            return t.from_json(d[t.reg_name])
    return d

def load(file: pathlib.Path):
    with open(file, 'r') as f:
        return json.load(f, object_hook=json_decoder)

def loads(payload: str):
    return json.loads(payload, object_hook=json_decoder)


@dataclasses.dataclass
class TypeEntry:
    type        : typing.Type
    reg_name    : str
    to_json     : typing.Callable
    from_json   : typing.Callable

TYPE_REGISTRY = []
def register_type(entry: TypeEntry):
    TYPE_REGISTRY.append(entry)
register_type(TypeEntry(pathlib.Path,'pathlib.Path',str,lambda x: pathlib.Path(x)))
register_type(TypeEntry(set,'builtin.set',list,lambda x: set(x)))
register_type(TypeEntry(tuple,'builtin.tuple',list,lambda x: tuple(x)))


import pathlib
import numpy as np
import pandas as pd
import math
import warnings

from . import DataQualityType
from .. import gaze_worldref, naming, pose, transforms

def _get_fields_for_dq_type(dq_type: DataQualityType) -> list[str|None]|None:
    match dq_type:
        case DataQualityType.viewpos_vidpos_homography:
            # from camera perspective, using homography
            # viewpos_vidpos_homography: using assumed viewing distance
            fields = [None, None, 'gazePosPlane2D_vidPos_homography']
        case DataQualityType.pose_vidpos_homography:
            # from camera perspective, using homography
            # pose_vidpos_homography   : using pose info
            fields = [None, 'gazePosCam_vidPos_homography', 'gazePosPlane2D_vidPos_homography']
        case DataQualityType.pose_vidpos_ray:
            # from camera perspective, using 3D gaze point ray
            fields = [None, 'gazePosCam_vidPos_ray', 'gazePosPlane2D_vidPos_ray']
        case DataQualityType.pose_world_eye:
            # using 3D world gaze position, with respect to eye tracker reference frame's origin
            fields = [None, 'gazePosCamWorld', 'gazePosPlane2DWorld']
        case DataQualityType.pose_left_eye:
            fields = ['gazeOriCamLeft', 'gazePosCamLeft', 'gazePosPlane2DLeft']
        case DataQualityType.pose_right_eye:
            fields = ['gazeOriCamRight', 'gazePosCamRight', 'gazePosPlane2DRight']
        case DataQualityType.pose_left_right_avg:
            fields = None   # special case, need to separately check availability of left and right
        case _:
            raise NotImplementedError(f'Logic for data quality type {dq} not implemented. Contact developer.')
    return fields

def compute(
        gazes: str|pathlib.Path|dict[int, list[gaze_worldref.Gaze]],
        poses: str|pathlib.Path|dict[int, pose.Pose],
        marker_intervals: str|pathlib.Path|pd.DataFrame,
        validation_intervals: list[list[int]],
        targets: dict[int, np.ndarray],
        distance_mm_for_homography: float,
        output_directory: str|pathlib.Path,
        filename: str = naming.validation_offset_fname,
        dq_types: list[DataQualityType]=None, allow_dq_fallback=False, include_data_loss=False
    ):
    output_directory = pathlib.Path(output_directory)
    if dq_types is None:
        dq_types = []

    # read input if needed
    if not isinstance(marker_intervals, pd.DataFrame):
        marker_intervals = pd.read_csv(marker_intervals, delimiter='\t', dtype={'marker_interval':int},index_col=['marker_interval','target'])
    if not all((c in marker_intervals or c in marker_intervals.index.names for c in ('marker_interval','target','start_timestamp','end_timestamp'))):
        raise ValueError('Provided marker intervals should contain the following columns: [marker_interval, target, start_timestamp, end_timestamp]')
    if not isinstance(gazes, dict):
        gazes = gaze_worldref.read_dict_from_file(gazes,validation_intervals)
    if not isinstance(poses, dict):
        poses = pose.read_dict_from_file(poses,validation_intervals)
    # prep targets
    targets_for_homography = {t_id: np.append(targets[t_id][0:2], distance_mm_for_homography) for t_id in targets}

    # for output
    out_df: pd.DataFrame = None

    # for each frame during analysis interval, determine offset
    # (angle) of gaze (each eye) to each of the targets
    for idx,iv in enumerate(validation_intervals):
        samples_per_frame = {k:v for (k,v) in gazes.items() if k>=iv[0] and k<=iv[1]}
        if not samples_per_frame:
            raise RuntimeError(f'There is no gaze data on the glassesValidator surface for validation interval (frames {idx[0]} to {idx[1]}), cannot proceed. This may be because there was no gaze during this interval or because the plane was not detected.')

        # check what data quality types we should output. Go with good defaults
        # first see what we have available
        dq_have: list[DataQualityType] = []
        for dq in DataQualityType:
            fields = _get_fields_for_dq_type(dq)
            if fields is None:
                continue
            have_data = np.vstack(tuple([not np.any(np.isnan(getattr(s,f))) for v in samples_per_frame.values() for s in v] for f in fields if f is not None))
            if np.any(np.all(have_data,axis=0)):
                dq_have.append(dq)

        if (DataQualityType.pose_left_eye in dq_have) and (DataQualityType.pose_right_eye in dq_have):
            dq_have.append(DataQualityType.pose_left_right_avg)
        # then determine, based on what user requests, what we will output
        if dq_types:
            if isinstance(dq_types,DataQualityType) or isinstance(dq_types, str):
                dq_types = [dq_types]
            else:
                # ensure list
                dq_types = list(dq_types)
            # do some checks on user input
            for i,dq in reversed(list(enumerate(dq_types))):
                if not isinstance(dq, DataQualityType):
                    if isinstance(dq, str):
                        if hasattr(DataQualityType, dq):
                            dq = dq_types[i] = getattr(DataQualityType, dq)
                        else:
                            raise ValueError(f"The string '{dq}' is not a known data quality type. Known types: {[e.name for e in DataQualityType]}")
                    else:
                        raise ValueError(f"The variable 'dq' should be a string with one of the following values: {[e.name for e in DataQualityType]}")
                if not dq in dq_have:
                    if allow_dq_fallback:
                        del dq_types[i]
                    else:
                        raise RuntimeError(f'Data quality type {dq} could not be used as its not available for this recording. Available data quality types: {[e.name for e in dq_have]}')

            if DataQualityType.pose_left_right_avg in dq_types:
                if (not DataQualityType.pose_left_eye in dq_have) or (not DataQualityType.pose_right_eye in dq_have):
                    if allow_dq_fallback:
                        dq_types.remove(DataQualityType.pose_left_right_avg)
                    else:
                        raise RuntimeError(f'Cannot use the data quality type {DataQualityType.pose_left_right_avg} because it requires having data quality types {DataQualityType.pose_left_eye} and {DataQualityType.pose_right_eye} available, but one or both are not available. Available data quality types: {[e.name for e in dq_have]}')

        if not dq_types:
            if DataQualityType.pose_vidpos_ray in dq_have:
                # highest priority is DataQualityType.pose_vidpos_ray
                dq_types.append(DataQualityType.pose_vidpos_ray)
            elif DataQualityType.pose_vidpos_homography in dq_have:
                # else at least try to use pose (shouldn't occur, if we have pose we have a calibrated camera, which means we should have the above)
                dq_types.append(DataQualityType.pose_vidpos_homography)
            else:
                # else we're down to falling back on an assumed viewing distance
                if not DataQualityType.viewpos_vidpos_homography in dq_have:
                    raise RuntimeError(f'Even data quality type {DataQualityType.viewpos_vidpos_homography} could not be used, bare minimum failed for some weird reason. Contact developer.')
                dq_types.append(DataQualityType.viewpos_vidpos_homography)

        # prepare output data frame
        df_idx  = []
        idxer = pd.IndexSlice
        idxs = marker_intervals.loc[idx+1,:].index.to_frame().to_numpy()
        for dq in dq_types:
            df_idx.append(np.vstack((np.full((idxs.shape[0],),idx+1,dtype='int'),idxs.shape[0]*[dq],idxs[:,0])).T)
        df_idx = pd.DataFrame(np.vstack(tuple(df_idx)),columns=[marker_intervals.index.names[0],'type',marker_intervals.index.names[1]])
        df  = pd.DataFrame(index=pd.MultiIndex.from_frame(df_idx.astype({marker_intervals.index.names[0]: 'int64', 'type': 'category', marker_intervals.index.names[1]: 'int64'})))

        # determine order in which targets were looked at
        for dq in dq_types:
            df.loc[idxer[idx+1,dq,:],'order'] = np.argsort(marker_intervals.loc(axis=0)[idx+1,:]['start_timestamp'].to_numpy())+1
        if df['order'].dtype==np.dtype('float64'):  # ensure int if needed
            df['order'] = df['order'].astype(np.int64)

        # now compute
        for t in targets:
            if (idx+1,t) not in marker_intervals.index:
                continue
            frame_idxs  =           [k           for k,v in samples_per_frame.items() for _ in v]
            ts          = np.vstack([s.timestamp for  v in samples_per_frame.values() for s in v])
            st = marker_intervals.loc[(idx+1,t),'start_timestamp']
            et = marker_intervals.loc[(idx+1,t),  'end_timestamp']
            q_data = np.logical_and(ts>=st, ts<=et)

            offset = np.full((np.count_nonzero(q_data),len(dq_types),2), np.nan)
            target_cam: dict[int,np.ndarray] = {}
            for idq,dq in enumerate(dq_types):
                # get data
                fields = _get_fields_for_dq_type(dq)
                if fields is None:
                    continue
                if fields[0] is None:
                    ori     = np.zeros((ts.shape[0],3))
                else:
                    ori     = np.vstack([getattr(s,fields[0]) for v in samples_per_frame.values() for s in v])
                gazePlane   = np.vstack([getattr(s,fields[2]) for v in samples_per_frame.values() for s in v])
                if fields[1] is None:
                    if not dq==DataQualityType.viewpos_vidpos_homography:
                        raise NotImplementedError(f'This field should be set, is a special case not implemented? Contact developer')
                    gaze    = np.hstack((gazePlane[:,0:2], np.full((gazePlane.shape[0],1),distance_mm_for_homography)))
                else:
                    gaze    = np.vstack([getattr(s,fields[1]) for v in samples_per_frame.values() for s in v])

                # compute
                out_idx = -1
                for i in range(len(ts)):
                    if not q_data[i]:
                        continue
                    out_idx += 1
                    frame_idx = frame_idxs[i]
                    if dq==DataQualityType.viewpos_vidpos_homography:
                        # get vectors based on assumed viewing distance (from config), without using pose info
                        vGaze   = gaze[i,:]
                        vTarget = targets_for_homography[t]
                    else:
                        # use 3D vectors known given pose information
                        if frame_idx not in poses:
                            continue
                        if frame_idx not in target_cam:
                            target_cam[frame_idx] = poses[frame_idx].world_frame_to_cam(targets[t])

                        # get vectors from origin to target and to gaze point
                        vGaze   = gaze[i,:]            -ori[i,:]
                        vTarget = target_cam[frame_idx]-ori[i,:]

                    # get offset
                    ang2D               = transforms.angle_between(vTarget,vGaze)
                    # decompose in horizontal/vertical (in plane space)
                    onPlaneAngle        = math.atan2(gazePlane[i,1]-targets[t][1], gazePlane[i,0]-targets[t][0])
                    offset[out_idx,idq,:]= ang2D*np.array([math.cos(onPlaneAngle), math.sin(onPlaneAngle)])

            # special case for average of left and right eye
            if DataQualityType.pose_left_right_avg in dq_types:
                l_idx = dq_types.index(DataQualityType.pose_left_eye)
                r_idx = dq_types.index(DataQualityType.pose_right_eye)
                a_idx = dq_types.index(DataQualityType.pose_left_right_avg)
                offset[:,a_idx,:] = offset[:,[l_idx, r_idx],:].mean(axis=1)

            # compute data quality for this target
            with warnings.catch_warnings():
                warnings.simplefilter("ignore") # ignore warnings from np.nanmean, np.nanmedian and np.nanstd

                df.loc[idxer[idx+1,:,t],'acc_x'] = np.nanmedian(offset[:,:,0],axis=0)
                df.loc[idxer[idx+1,:,t],'acc_y'] = np.nanmedian(offset[:,:,1],axis=0)
                # NB: below is equivalent to computing offset of median gaze vector from target
                df.loc[idxer[idx+1,:,t],'acc'  ] = np.hypot(df.loc[idxer[idx+1,:,t],'acc_x'], df.loc[idxer[idx+1,:,t],'acc_y'])

                df.loc[idxer[idx+1,:,t],'rms_x'] = np.sqrt(np.nanmean(np.diff(offset[:,:,0],axis=0)**2))
                df.loc[idxer[idx+1,:,t],'rms_y'] = np.sqrt(np.nanmean(np.diff(offset[:,:,1],axis=0)**2))
                # NB: below is equivalent to np.sqrt(np.nanmean(xdiff+ydiff))
                df.loc[idxer[idx+1,:,t],'rms'  ] = np.hypot(df.loc[idxer[idx+1,:,t],'rms_x'], df.loc[idxer[idx+1,:,t],'rms_y'])

                df.loc[idxer[idx+1,:,t],'std_x'] = np.nanstd(offset[:,:,0],ddof=0,axis=0)
                df.loc[idxer[idx+1,:,t],'std_y'] = np.nanstd(offset[:,:,1],ddof=0,axis=0)
                df.loc[idxer[idx+1,:,t],'std'  ] = np.hypot(df.loc[idxer[idx+1,:,t],'std_x'], df.loc[idxer[idx+1,:,t],'std_y'])

                if include_data_loss:
                    df.loc[idxer[idx+1,:,t],'data_loss'] = np.sum(np.isnan(offset[:,:,0]),axis=0)/offset.shape[0]

        if out_df is None:
            out_df = df
        else:
            out_df = pd.concat((out_df,df))

    out_df.to_csv(output_directory / filename, mode='w', header=True, sep='\t', na_rep='nan', float_format="%.6f")